import numpy as np

from gtnn.network.activation import LogSigmoid
from gtnn.network.net import Net


def mlp(sizes,
        weightGenerator=np.random.random,
        biasGenerator=np.random.random,
        activationFunction=LogSigmoid(-1, 1)):
    r"""Geneerate Multilayer Perceptron (MLP) and return as a Net

    Parameters
    ----------
    sizes : list of integers
        Size of each layer
    weightGenerator : functor
        Functor for generating weights
    biasGenerator : functor
        Functor for generating bias values
    activationFunction : activation function
        Activation function - the same for all neurons

    Returns
    -------
    net : :class:`~gtnn.network.Net`
        Net object with the requested architecture

    Notes
    -----
    Creates MLP

    """
    n = Net(sizes[0], sizes[-1])
    layerId = n.addVertexProperty("layerId", "short")
    lastLayer = []
    presentLayer = []

    # Create all layers
    for layerIdx, size in enumerate(sizes):
        for i in range(size):
            v = n.g.add_vertex()
            layerId[v] = layerIdx
            n.activation[v] = activationFunction
            n.biasProp[v] = biasGenerator()

            presentLayer.append(v)
            for l in lastLayer:
                e = n.g.add_edge(l, v)
                n.weightProp[e] = weightGenerator()

        lastLayer = list(presentLayer)
        presentLayer = list()

    n.prepare()
    return n



import numpy as np


def train(net, inputs, outputs, numEpochs=100, learningRate=0.1):
    inputs = np.array(inputs)
    outputs = np.array(outputs)

    epochIdx = 0
    while True:
        for inp, out in zip(inputs, outputs):
            netOut = net.forward(inp)
            err = netOut - out
            net.backward(err)

            # Weights learning:
            for e in net.g.edges():
                net.weightProp[e] -= net.valueProp[e.source()] * \
                    learningRate * net.errorProp[e.target()]

            # Bias learning
            for v in net.g.vertices():
                net.biasProp[v] -= learningRate * net.errorProp[v]

        # End condition
        epochIdx += 1
        if epochIdx > numEpochs:
            break



import numpy as np


def train(net, inputs, outputs, numEpochs=100, learningRate=0.1, momentum=0.6):
    inputs = np.array(inputs)
    outputs = np.array(outputs)

    prevWeightChangeProp = net.addEdgeProperty("previous weight change",
                                               "long double")

    prevBiasChangeProp = net.addVertexProperty("previous bias change",
                                               "long double")
    epochIdx = 0
    while True:
        for inp, out in zip(inputs, outputs):
            netOut = net.forward(inp)
            err = netOut - out
            net.backward(err)

            # Weights learning:
            for e in net.g.edges():
                prevWeightChangeProp[e] = -net.valueProp[e.source()] * \
                    learningRate * net.errorProp[e.target()] + \
                    prevWeightChangeProp[e] * momentum

                net.weightProp[e] += prevWeightChangeProp[e]

            # Bias learning
            for v in net.g.vertices():
                prevBiasChangeProp[v] = -learningRate * net.errorProp[v] + \
                    momentum * prevBiasChangeProp[v]

                net.biasProp[v] += prevBiasChangeProp[v]

        # End condition
        epochIdx += 1
        if epochIdx > numEpochs:
            break
